@@ -6,9 +6,6 @@
  * Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
  */
 
-#include <sstream>
-#include <iostream>
-#include <iomanip>
 #include <cctype>
 
 #include <qpdf/Constants.h>
@@ -28,6 +25,8 @@
 
 #include "pikepdf.h"
 
+#include "object_parsers.h"
+
 /*
 Type table
 
@@ -54,126 +53,6 @@ needed.
 
 */
 
-class PyParserCallbacks : public QPDFObjectHandle::ParserCallbacks {
-public:
-    using QPDFObjectHandle::ParserCallbacks::ParserCallbacks;
-    virtual ~PyParserCallbacks() = default;
-
-    void handleObject(QPDFObjectHandle h) override {
-        PYBIND11_OVERLOAD_PURE_NAME(
-            void,
-            QPDFObjectHandle::ParserCallbacks,
-            "handle_object", /* Python name */
-            handleObject, /* C++ name */
-            h
-        );
-    }
-
-    void handleEOF() override {
-        PYBIND11_OVERLOAD_PURE_NAME(
-            void,
-            QPDFObjectHandle::ParserCallbacks,
-            "handle_eof", /* Python name */
-            handleEOF, /* C++ name; trailing comma needed for macro */
-        );
-    }
-};
-
-
-class OperandGrouper : public QPDFObjectHandle::ParserCallbacks {
-public:
-    OperandGrouper(const std::string& operators)
-        : parsing_inline_image(false), count(0)
-    {
-        std::istringstream f(operators);
-        std::string s;
-        while (std::getline(f, s, ' ')) {
-            this->whitelist.insert(s);
-        }
-    }
-
-    void handleObject(QPDFObjectHandle obj) override
-    {
-        this->count++;
-        if (obj.getTypeCode() == QPDFObject::object_type_e::ot_operator) {
-            std::string op = obj.getOperatorValue();
-
-            // If we have a whitelist and this operator is not on the whitelist,
-            // discard it and all the tokens we collected
-            if (!this->whitelist.empty()) {
-                if (op[0] == 'q' || op[0] == 'Q') {
-                    // We have token with multiple stack push/pops
-                    if (this->whitelist.count("q") == 0 && this->whitelist.count("Q") == 0) {
-                        this->tokens.clear();
-                        return;
-                    }
-                } else if (this->whitelist.count(op) == 0) {
-                    this->tokens.clear();
-                    return;
-                }
-            }
-            if (op == "BI") {
-                this->parsing_inline_image = true;
-            } else if (this->parsing_inline_image) {
-                if (op == "ID") {
-                    this->inline_metadata = this->tokens;
-                } else if (op == "EI") {
-                    auto PdfInlineImage = py::module::import("pikepdf").attr("PdfInlineImage");
-                    auto kwargs = py::dict();
-                    kwargs["image_data"] = this->tokens.at(0);
-                    kwargs["image_object"] = this->inline_metadata;
-                    auto iimage = PdfInlineImage(**kwargs);
-
-                    // Package as list with single element for consistency
-                    auto iimage_list = py::list();
-                    iimage_list.append(iimage);
-
-                    auto instruction = py::make_tuple(
-                        iimage_list,
-                        QPDFObjectHandle::newOperator("INLINE IMAGE")
-                    );
-                    this->instructions.append(instruction);
-
-                    this->parsing_inline_image = false;
-                    this->inline_metadata.clear();
-                }
-            } else {
-                py::list operand_list = py::cast(this->tokens);
-                auto instruction = py::make_tuple(operand_list, obj);
-                this->instructions.append(instruction);
-            }
-            this->tokens.clear();
-        } else {
-            this->tokens.push_back(obj);
-        }
-    }
-
-    void handleEOF() override
-    {
-        if (!this->tokens.empty())
-            this->warning = "Unexpected end of stream";
-    }
-
-    py::list getInstructions() const
-    {
-        return this->instructions;
-    }
-
-    std::string getWarning() const
-    {
-        return this->warning;
-    }
-
-private:
-    std::set<std::string> whitelist;
-    std::vector<QPDFObjectHandle> tokens;
-    bool parsing_inline_image;
-    std::vector<QPDFObjectHandle> inline_metadata;
-    py::list instructions;
-    uint count;
-    std::string warning;
-};
-
 
 size_t list_range_check(QPDFObjectHandle h, int index)
 {
@@ -363,7 +242,8 @@ void init_object(py::module& m)
             [](QPDFObjectHandle &h, std::shared_ptr<QPDF> possible_owner) {
                 return (h.getOwningQPDF() == possible_owner.get());
             },
-            "Test if this object is owned by the indicated *possible_owner*."
+            "Test if this object is owned by the indicated *possible_owner*.",
+            py::arg("possible_owner")
         )
         .def_property_readonly("is_indirect", &QPDFObjectHandle::isIndirect)
         .def("__repr__", &objecthandle_repr)
@@ -373,8 +253,6 @@ void init_object(py::module& m)
 
                 //Objects which compare equal must have the same hash value
                 switch (self.getTypeCode()) {
-                    case QPDFObject::object_type_e::ot_null:
-                        return py::int_(0);
                     case QPDFObject::object_type_e::ot_string:
                     {
                         return hash(py::bytes(self.getUTF8Value()));
@@ -387,7 +265,7 @@ void init_object(py::module& m)
                     case QPDFObject::object_type_e::ot_dictionary:
                     case QPDFObject::object_type_e::ot_stream:
                     case QPDFObject::object_type_e::ot_inlineimage:
-                        throw py::value_error("Can't hash mutable object");
+                        throw py::type_error("Can't hash mutable object");
                     default:
                         break;
                 }
@@ -412,26 +290,13 @@ void init_object(py::module& m)
                 }
             }
         )
-        .def("__eq__",
-            [](QPDFObjectHandle &self, py::bytes other) {
-                std::string bytes_other = other.cast<std::string>();
-                switch (self.getTypeCode()) {
-                    case QPDFObject::object_type_e::ot_string:
-                        return self.getStringValue() == bytes_other;
-                    case QPDFObject::object_type_e::ot_name:
-                        return self.getName() == bytes_other;
-                    default:
-                        return false;
-                }
-            }
-        )
         .def("__eq__",
             [](QPDFObjectHandle &self, py::object other) -> py::object {
                 QPDFObjectHandle q_other;
                 try {
                     q_other = objecthandle_encode(other);
                 } catch (const py::cast_error&) {
-                    return py::globals()["__builtins__"].attr("NotImplemented");
+                    return py::globals()["__builtins__"]["NotImplemented"];
                 }
                 bool result = (self == objecthandle_encode(other));
                 return py::bool_(result);
@@ -443,7 +308,7 @@ void init_object(py::module& m)
                     return (Py_ssize_t)h.getDictAsMap().size(); // getKeys constructs a new object, so this is better
                 else if (h.isArray())
                     return (Py_ssize_t)h.getArrayNItems();
-                throw py::value_error("length not defined for object");
+                throw py::type_error("length not defined for object");
             }
         )
         .def("__getitem__",
@@ -550,14 +415,14 @@ void init_object(py::module& m)
                 QPDFObjectHandle value;
                 try {
                     value = object_get_key(h, key);
-                } catch (const py::key_error &e) {
+                } catch (const py::key_error&) {
                     return default_;
                 }
                 return py::cast(value);
             },
-            "for dictionary objects, behave as dict.get(key, default=None)",
+            "For ``pikepdf.Dictionary`` objects, behave as ``dict.get(key, default=None)``",
             py::arg("key"),
-            py::arg("default_") = py::none(),
+            py::arg("default") = py::none(),
             py::return_value_policy::reference_internal
         )
         .def("get",
@@ -565,14 +430,14 @@ void init_object(py::module& m)
                 QPDFObjectHandle value;
                 try {
                     value = object_get_key(h, name.getName());
-                } catch (const py::key_error &e) {
+                } catch (const py::key_error&) {
                     return default_;
                 }
                 return py::cast(value);
             },
-            "for dictionary objects, behave as dict.get(key, default=None)",
+            "For ``pikepdf.Dictionary`` objects, behave as ``dict.get(key, default=None)``",
             py::arg("key"),
-            py::arg("default_") = py::none(),
+            py::arg("default") = py::none(),
             py::return_value_policy::reference_internal
         )
         .def("keys", &QPDFObjectHandle::getKeys)
@@ -664,6 +529,15 @@ void init_object(py::module& m)
                 h.eraseItem(u_index);
             }
         )
+        .def("wrap_in_array",
+            [](QPDFObjectHandle &h) {
+                return h.wrapInArray();
+            },
+            "Return the object wrapped in an array if not already an array."
+        )
+        .def_property_readonly("is_rectangle", &QPDFObjectHandle::isRectangle,
+            "Returns True if the object is a rectangle (an array of 4 numbers)"
+        )
         .def("get_stream_buffer",
             [](QPDFObjectHandle &h) {
                 PointerHolder<Buffer> phbuf = h.getStreamData();
@@ -694,37 +568,17 @@ void init_object(py::module& m)
             },
             "Read the content stream associated with this object without decoding"
         )
-        .def("write",
-            [](QPDFObjectHandle &h, py::bytes data, py::args args, py::kwargs kwargs) {
+        .def("_write",
+            [](QPDFObjectHandle &h, py::bytes data, py::object filter, py::object decode_parms) {
                 std::string sdata = data;
-                QPDFObjectHandle filter = QPDFObjectHandle::newNull();
-                QPDFObjectHandle decode_parms = QPDFObjectHandle::newNull();
-                if (args.size() != 0)
-                    throw py::value_error("Too many positional arguments");
-                if (kwargs.contains("filter"))
-                    filter = objecthandle_encode(kwargs["filter"]);
-                if (kwargs.contains("decode_parms"))
-                    decode_parms = objecthandle_encode(kwargs["decode_parms"]);
-                h.replaceStreamData(sdata, filter, decode_parms);
+                QPDFObjectHandle h_filter = objecthandle_encode(filter);
+                QPDFObjectHandle h_decode_parms = objecthandle_encode(decode_parms);
+                h.replaceStreamData(sdata, h_filter, h_decode_parms);
             },
             R"~~~(
-            Replace the content stream with `data`, compressed according to `filter` and `decode_parms`
-
-            :param data: the new data to use for replacement
-            :type data: bytes
-            :param filter: The filter(s) with which the data is (already) encoded
-            :param decode_parms: Parameters for the filters with which the object is encode
-
-            If only one `filter` is specified, it may be a name such as
-            `Name('/FlateDecode')`. If there are multiple filters, then array
-            of names should be given.
-
-            If there is only one filter, `decode_parms` is a Dictionary of
-            parameters for that filter. If there are multiple filters, then
-            `decode_parms` is an Array of Dictionary, where each array index
-            is corresponds to the filter.
-
-            )~~~"
+            Low level write/replace stream data without argument checking. Use .write().
+            )~~~",
+            py::arg("data"), py::arg("filter"), py::arg("decode_parms")
         )
         .def_property_readonly("images",
             [](QPDFObjectHandle &h) {
@@ -749,7 +603,16 @@ void init_object(py::module& m)
             py::arg("prepend") = false,
             py::keep_alive<1, 2>()
         )
-        .def("page_contents_coalesce", &QPDFObjectHandle::coalesceContentStreams)
+        .def("page_contents_coalesce", &QPDFObjectHandle::coalesceContentStreams,
+            R"~~~(
+            Coalesce an array of page content streams into a single content stream.
+
+            The PDF specification allows the ``/Contents`` object to contain either
+            an array of content streams or a single content stream. However, it
+            simplifies parsing and editing if there is only a single content stream.
+            This function merges all content streams.
+            )~~~"
+        )
         .def_property_readonly("_objgen",
             &object_get_objgen
         )
@@ -811,6 +674,41 @@ void init_object(py::module& m)
             py::arg("resolved") = false,
             "Convert PDF objects into their binary representation, optionally resolving indirect objects."
         )
+        .def("to_json",
+            [](QPDFObjectHandle &h, bool dereference = false) -> py::bytes {
+                return h.getJSON(dereference).unparse();
+            },
+            py::arg("dereference") = false,
+            R"~~~(
+            Convert to a QPDF JSON representation of the object.
+
+            See the QPDF manual for a description of its JSON representation.
+            http://qpdf.sourceforge.net/files/qpdf-manual.html#ref.json
+
+            Not necessarily compatible with other PDF-JSON representations that
+            exist in the wild.
+
+            * Names are encoded as UTF-8 strings
+            * Indirect references are encoded as strings containing ``obj gen R``
+            * Strings are encoded as UTF-8 strings with unrepresentable binary
+              characters encoded as ``\uHHHH``
+            * Encoding streams just encodes the stream's dictionary; the stream
+              data is not represented
+            * Object types that are only valid in content streams (inline
+              image, operator) as well as "reserved" objects are not
+              representable and will be serialized as ``null``.
+
+            Args:
+                dereference (bool): If True, dereference the object is this is an
+                    indirect object.
+
+            Returns:
+                bytes: JSON bytestring of object. The object is UTF-8 encoded
+                and may be decoded to a Python str that represents the binary
+                values ``\x00-\xFF`` as ``U+0000`` to ``U+00FF``; that is,
+                it may contain mojibake.
+            )~~~"
+        )
         ; // end of QPDFObjectHandle bindings
 
     m.def("_new_boolean", &QPDFObjectHandle::newBool, "Construct a PDF Boolean object");
@@ -831,10 +729,10 @@ void init_object(py::module& m)
     );
     m.def("_new_name",
         [](const std::string& s) {
-            if (s.at(0) != '/')
-                throw py::value_error("Name objects must begin with '/'");
             if (s.length() < 2)
                 throw py::value_error("Name must be at least one character long");
+            if (s.at(0) != '/')
+                throw py::value_error("Name objects must begin with '/'");
             return QPDFObjectHandle::newName(s);
         },
         "Create a Name from a string. Must begin with '/'. All other characters except null are valid."
@@ -900,17 +798,22 @@ void init_object(py::module& m)
         [](const std::string& op) {
             return QPDFObjectHandle::newOperator(op);
         },
-        "Construct a PDF Operator object for use in content streams"
+        "Construct a PDF Operator object for use in content streams.",
+        py::arg("op")
     );
     m.def("_Null", &QPDFObjectHandle::newNull,
         "Construct a PDF Null object"
     );
 
-    py::class_<QPDFObjectHandle::ParserCallbacks, PyParserCallbacks> parsercallbacks(m, "StreamParser");
-    parsercallbacks
+    py::class_<QPDFObjectHandle::ParserCallbacks, PyParserCallbacks>(m, "StreamParser")
         .def(py::init<>())
-        .def("handle_object", &QPDFObjectHandle::ParserCallbacks::handleObject)
-        .def("handle_eof", &QPDFObjectHandle::ParserCallbacks::handleEOF);
+        .def("handle_object",
+            [](QPDFObjectHandle::ParserCallbacks &parsercallbacks, QPDFObjectHandle &h) {
+                parsercallbacks.handleObject(h);
+            }
+        )
+        .def("handle_eof", &QPDFObjectHandle::ParserCallbacks::handleEOF)
+        ;
 
     m.def("_encode",
         [](py::handle handle) {
