@@ -28,10 +28,9 @@
 
 #include "qpdf_pagelist.h"
 #include "qpdf_inputsource.h"
+#include "pipeline.h"
 #include "utils.h"
 
-extern "C" const char* qpdf_get_qpdf_version();
-
 
 void check_stream_is_usable(py::object stream)
 {
@@ -58,6 +57,7 @@ open_pdf(
     q->setPasswordIsHexKey(hex_password);
     q->setIgnoreXRefStreams(ignore_xref_streams);
     q->setAttemptRecovery(attempt_recovery);
+    q->setImmediateCopyFrom(true);
 
     if (py::hasattr(filename_or_stream, "read") && py::hasattr(filename_or_stream, "seek")) {
         // Python code gave us an object with a stream interface
@@ -66,7 +66,7 @@ open_pdf(
         check_stream_is_usable(stream);
 
         // The PythonInputSource object will be owned by q
-        InputSource* input_source = new PythonInputSource(stream);
+        auto input_source = PointerHolder<InputSource>(new PythonInputSource(stream));
         py::gil_scoped_release release;
         q->processInputSource(input_source, password.c_str());
     } else {
@@ -80,7 +80,7 @@ open_pdf(
         q->processFile(
             description.c_str(),
             file, // transferring ownership
-            true, // QPDF will close the file
+            true, // QPDF will close the file (including if there are exceptions)
             password.c_str()
         );
         file = nullptr; // QPDF owns the file and will close it
@@ -116,38 +116,301 @@ private:
 };
 
 
+void update_xmp_pdfversion(QPDF &q, std::string version)
+{
+    auto impl = py::module::import("pikepdf._cpphelpers").attr("update_xmp_pdfversion");
+    auto pypdf = py::cast(q);
+    impl(pypdf, version);
+}
+
+
+void setup_encryption(
+    QPDFWriter &w,
+    py::object encryption,
+    std::string &owner,
+    std::string &user
+)
+{
+    bool aes = true;
+    bool metadata = true;
+    std::map<std::string, bool> allow;
+    int encryption_level = 6;
+
+    if (encryption.contains("R")) {
+        if (!py::isinstance<py::int_>(encryption["R"]))
+            throw py::type_error("Encryption level 'R' must be an integer");
+        encryption_level = py::int_(encryption["R"]);
+    }
+    if (encryption_level < 2 || encryption_level > 6)
+        throw py::value_error("Invalid encryption level: must be 2, 3, 4 or 6");
+
+    if (encryption_level == 5) {
+        auto warn = py::module::import("warnings").attr("warn");
+        warn("Encryption R=5 is deprecated");
+    }
+
+    if (encryption.contains("owner")) {
+        if (encryption_level <= 4) {
+            auto success = QUtil::utf8_to_pdf_doc(encryption["owner"].cast<std::string>(), owner);
+            if (!success)
+                throw py::value_error("Encryption level is R3/R4 and password is not encodable as PDFDocEncoding");
+        } else {
+            owner = encryption["owner"].cast<std::string>();
+        }
+    }
+    if (encryption.contains("user")) {
+        if (encryption_level <= 4) {
+            auto success = QUtil::utf8_to_pdf_doc(encryption["user"].cast<std::string>(), user);
+            if (!success)
+                throw py::value_error("Encryption level is R3/R4 and password is not encodable as PDFDocEncoding");
+        } else {
+            user = encryption["user"].cast<std::string>();
+        }
+    }
+    if (encryption.contains("allow")) {
+        auto pyallow = encryption["allow"];
+        allow["accessibility"] = pyallow.attr("accessibility").cast<bool>();
+        allow["extract"] = pyallow.attr("extract").cast<bool>();
+        allow["modify_assembly"] = pyallow.attr("modify_assembly").cast<bool>();
+        allow["modify_annotation"] = pyallow.attr("modify_annotation").cast<bool>();
+        allow["modify_form"] = pyallow.attr("modify_form").cast<bool>();
+        allow["modify_other"] = pyallow.attr("modify_other").cast<bool>();
+        allow["print_lowres"] = pyallow.attr("print_lowres").cast<bool>();
+        allow["print_highres"] = pyallow.attr("print_highres").cast<bool>();
+    }
+    if (encryption.contains("aes")) {
+        if (py::isinstance<py::bool_>(encryption["aes"]))
+            aes = py::bool_(encryption["aes"]);
+        else
+            throw py::type_error("aes must be bool");
+    } else {
+        aes = (encryption_level >= 4);
+    }
+    if (encryption.contains("metadata")) {
+        if (py::isinstance<py::bool_>(encryption["metadata"]))
+            metadata = py::bool_(encryption["metadata"]);
+        else
+            throw py::type_error("metadata must be bool");
+    } else {
+        metadata = (encryption_level >= 4);
+    }
+
+    if (metadata && encryption_level < 4) {
+        throw py::value_error("Cannot encrypt metadata when R < 4");
+    }
+    if (aes && encryption_level < 4) {
+        throw py::value_error("Cannot encrypt with AES when R < 4");
+    }
+    if (encryption_level == 6 && !aes) {
+        throw py::value_error("When R = 6, AES encryption must be enabled");
+    }
+    if (metadata && !aes) {
+        throw py::value_error("Cannot encrypt metadata unless AES encryption is enabled");
+    }
+
+    qpdf_r3_print_e print;
+    if (allow["print_highres"])
+        print = qpdf_r3p_full;
+    else if (allow["print_lowres"])
+        print = qpdf_r3p_low;
+    else
+        print = qpdf_r3p_none;
+
+    if (encryption_level == 6) {
+        w.setR6EncryptionParameters(
+            user.c_str(), owner.c_str(),
+            allow["accessibility"],
+            allow["extract"],
+            allow["modify_assembly"],
+            allow["modify_annotation"],
+            allow["modify_form"],
+            allow["modify_other"],
+            print,
+            metadata
+        );
+    } else if (encryption_level == 5) {
+        // TODO WARNING
+        w.setR5EncryptionParameters(
+            user.c_str(), owner.c_str(),
+            allow["accessibility"],
+            allow["extract"],
+            allow["modify_assembly"],
+            allow["modify_annotation"],
+            allow["modify_form"],
+            allow["modify_other"],
+            print,
+            metadata
+        );
+    } else if (encryption_level == 4) {
+        w.setR4EncryptionParameters(
+            user.c_str(), owner.c_str(),
+            allow["accessibility"],
+            allow["extract"],
+            allow["modify_assembly"],
+            allow["modify_annotation"],
+            allow["modify_form"],
+            allow["modify_other"],
+            print,
+            metadata,
+            aes
+        );
+    } else if (encryption_level == 3) {
+        w.setR3EncryptionParameters(
+            user.c_str(), owner.c_str(),
+            allow["accessibility"],
+            allow["extract"],
+            allow["modify_assembly"],
+            allow["modify_annotation"],
+            allow["modify_form"],
+            allow["modify_other"],
+            print
+        );
+    } else if (encryption_level == 2) {
+        w.setR2EncryptionParameters(
+            user.c_str(), owner.c_str(),
+            (print != qpdf_r3p_none),
+            allow["modify_assembly"],
+            allow["extract"],
+            allow["modify_annotation"]
+        );
+    }
+}
+
+
+typedef std::pair<std::string, int> pdf_version_extension;
+
+pdf_version_extension get_version_extension(py::object ver_ext)
+{
+    std::string version = "";
+    int extension = 0;
+    try {
+        version = ver_ext.cast<std::string>();
+        extension = 0;
+    } catch (const py::cast_error&) {
+        try {
+            auto version_ext = ver_ext.cast<pdf_version_extension>();
+            version = version_ext.first;
+            extension = version_ext.second;
+        } catch (const py::cast_error&) {
+            throw py::type_error("PDF version must be a tuple: (str, int)");
+        }
+    }
+    return pdf_version_extension(version, extension);
+}
+
+
+/* Helper class to ensure streams we open get closed by destructor */
+class Closer
+{
+public:
+    Closer() : monitored(py::none()) {}
+    ~Closer() {
+        if (!this->monitored.is_none()) {
+            this->monitored.attr("close")();
+        }
+    }
+    void set(py::object monitored) {
+        this->monitored = monitored;
+    }
+    Closer(const Closer& other) = delete;
+    Closer(Closer&& other) = delete;
+    Closer& operator= (const Closer& other) = delete;
+    Closer& operator= (Closer&& other) = delete;
+
+private:
+    py::object monitored;
+};
+
 void save_pdf(
     QPDF& q,
     py::object filename_or_stream,
     bool static_id=false,
     bool preserve_pdfa=true,
-    std::string min_version="",
-    std::string force_version="",
+    py::object min_version=py::none(),
+    py::object force_version=py::none(),
+    bool fix_metadata_version=true,
     bool compress_streams=true,
-    qpdf_stream_decode_level_e stream_decode_level=qpdf_dl_generalized,
+    py::object stream_decode_level=py::none(),
     qpdf_object_stream_e object_stream_mode=qpdf_o_preserve,
     bool normalize_content=false,
     bool linearize=false,
     bool qdf=false,
-    py::object progress=py::none())
+    py::object progress=py::none(),
+    py::object encryption=py::none())
 {
+    std::string owner;
+    std::string user;
+    std::string description;
     QPDFWriter w(q);
 
-    // Parameters
     if (static_id) {
         w.setStaticID(true);
     }
     w.setNewlineBeforeEndstream(preserve_pdfa);
-    if (!min_version.empty()) {
-        w.setMinimumPDFVersion(min_version, 0);
-    }
-    if (!force_version.empty()) {
-        w.forcePDFVersion(force_version, 0);
+
+    if (!min_version.is_none()) {
+        auto version_ext = get_version_extension(min_version);
+        w.setMinimumPDFVersion(version_ext.first, version_ext.second);
     }
     w.setCompressStreams(compress_streams);
-    w.setDecodeLevel(stream_decode_level);
+    if (!stream_decode_level.is_none()) {
+        // Unconditionally calling setDecodeLevel has side effects, disabling
+        // preserve encryption in particular
+        w.setDecodeLevel(stream_decode_level.cast<qpdf_stream_decode_level_e>());
+    }
     w.setObjectStreamMode(object_stream_mode);
 
+    py::object stream;
+    Closer stream_closer;
+
+    if (py::hasattr(filename_or_stream, "write") && py::hasattr(filename_or_stream, "seek")) {
+        // Python code gave us an object with a stream interface
+        stream = filename_or_stream;
+        check_stream_is_usable(stream);
+        description = py::repr(stream);
+    } else {
+        py::object filename = fspath(filename_or_stream);
+        py::object ospath = py::module::import("os").attr("path");
+        py::object samefile = ospath.attr("samefile");
+        try {
+            if (samefile(filename, q.getFilename()).cast<bool>()) {
+                throw py::value_error("Cannot overwrite input file");
+            }
+        } catch (const py::error_already_set &e) {
+            // We expect FileNotFoundError is filename refers to a file that does
+            // not exist, or if q.getFilename indicates a memory file. Suppress
+            // that, and rethrow all others.
+            if (!e.matches(PyExc_FileNotFoundError))
+                throw;
+        }
+        stream = py::module::import("io").attr("open")(filename, "wb");
+        stream_closer.set(stream);
+        description = py::str(filename);
+    }
+
+    // We must set up the output pipeline before we configure encryption
+    Pl_PythonOutput output_pipe(description.c_str(), stream);
+    w.setOutputPipeline(&output_pipe);
+
+    if (encryption.is(py::bool_(true)) && !q.isEncrypted()) {
+        throw py::value_error("can't perserve encryption parameters on a file with no encryption");
+    }
+
+    if (
+        (encryption.is(py::bool_(true)) || py::isinstance<py::dict>(encryption))
+            && (normalize_content || !stream_decode_level.is_none())
+    ) {
+        throw py::value_error("cannot save with encryption and normalize_content or stream_decode_level");
+    }
+
+    if (encryption.is(py::bool_(true))) {
+        w.setPreserveEncryption(true); // Keep existing encryption
+    } else if (encryption.is_none() || encryption.is(py::bool_(false))) {
+        w.setPreserveEncryption(false); // Remove encryption
+    } else {
+        setup_encryption(w, encryption, owner, user);
+    }
+
     if (normalize_content && linearize) {
         throw py::value_error("cannot save with both normalize_content and linearize");
     }
@@ -155,83 +418,25 @@ void save_pdf(
     w.setLinearization(linearize);
     w.setQDFMode(qdf);
 
+    if (!force_version.is_none()) {
+        auto version_ext = get_version_extension(force_version);
+        w.forcePDFVersion(version_ext.first, version_ext.second);
+    }
+    if (fix_metadata_version) {
+        update_xmp_pdfversion(q, w.getFinalVersion());
+    }
+
     if (!progress.is_none()) {
         auto reporter = PointerHolder<QPDFWriter::ProgressReporter>(new PikeProgressReporter(progress));
         w.registerProgressReporter(reporter);
     }
 
-    if (py::hasattr(filename_or_stream, "write") && py::hasattr(filename_or_stream, "seek")) {
-        // Python code gave us an object with a stream interface
-        py::object stream = filename_or_stream;
-        check_stream_is_usable(stream);
-
-        // TODO might be able to improve this by streaming rather than buffering
-        // using subclass of Pipeline that routes calls to Python.
-        w.setOutputMemory();
-
-        // It would be kind to release the GIL here, but this is not possible if
-        // another thread has an object and tries to mess with it. Correctness
-        // is more important than performance.
-        w.write();
-
-        // But now that we've held the GIL forever, we can release it and take
-        // it back again; at least in theory giving other threads a chance to
-        // to do something.
-        {
-            py::gil_scoped_release release;
-        }
-
-        // getBuffer returns Buffer* and qpdf says we are responsible for
-        // deleting it, so capture it in a unique_ptr
-        std::unique_ptr<Buffer> output_buffer(w.getBuffer());
-
-        // Create a memoryview of the buffer that libqpdf created
-        // Awkward API alert:
-        //     QPDFWriter::getBuffer -> Buffer*  (caller frees memory)
-        // and  Buffer::getBuffer -> unsigned char*  (caller does not own memory)
-        py::buffer_info output_buffer_info(
-            output_buffer->getBuffer(),
-            output_buffer->getSize());
-        py::memoryview view_output_buffer(output_buffer_info);
-
-        // Send it to the stream object (probably copying)
-        stream.attr("write")(view_output_buffer);
-    } else {
-        py::object filename = filename_or_stream;
-        std::string description = py::str(filename);
-        // Delete the intended filename, in case it is the same as the input file.
-        // This ensures that the input file will continue to exist in memory on Linux.
-        portable_unlink(filename);
-        FILE* file = portable_fopen(filename, "wb");
-        w.setOutputFile(description.c_str(), file, true);
-        w.write();
-        file = nullptr; // QPDF will close it
-    }
+    w.write();
 }
 
 
-PYBIND11_MODULE(_qpdf, m) {
-    //py::options options;
-    //options.disable_function_signatures();
-
-    m.doc() = "pikepdf provides a Pythonic interface for QPDF";
-
-    m.def("qpdf_version", &qpdf_get_qpdf_version, "Get libqpdf version");
-
-    static py::exception<QPDFExc> exc_main(m, "PdfError");
-    static py::exception<QPDFExc> exc_password(m, "PasswordError");
-    py::register_exception_translator([](std::exception_ptr p) {
-        try {
-            if (p) std::rethrow_exception(p);
-        } catch (const QPDFExc &e) {
-            if (e.getErrorCode() == qpdf_e_password) {
-                exc_password(e.what());
-            } else {
-                exc_main(e.what());
-            }
-        }
-    });
-
+void init_qpdf(py::module &m)
+{
     py::enum_<qpdf_object_stream_e>(m, "ObjectStreamMode")
         .value("disable", qpdf_object_stream_e::qpdf_o_disable)
         .value("preserve", qpdf_object_stream_e::qpdf_o_preserve)
@@ -243,7 +448,12 @@ PYBIND11_MODULE(_qpdf, m) {
         .value("specialized", qpdf_stream_decode_level_e::qpdf_dl_specialized)
         .value("all", qpdf_stream_decode_level_e::qpdf_dl_all);
 
-    init_pagelist(m);
+    py::enum_<QPDF::encryption_method_e>(m, "EncryptionMethod")
+        .value("none", QPDF::encryption_method_e::e_none)
+        .value("unknown", QPDF::encryption_method_e::e_unknown)
+        .value("rc4", QPDF::encryption_method_e::e_rc4)
+        .value("aes", QPDF::encryption_method_e::e_aes)
+        .value("aesv3", QPDF::encryption_method_e::e_aesv3);
 
     py::class_<QPDF, std::shared_ptr<QPDF>>(m, "Pdf", "In-memory representation of a PDF")
         .def_static("new",
@@ -253,20 +463,31 @@ PYBIND11_MODULE(_qpdf, m) {
                 q->setSuppressWarnings(true);
                 return q;
             },
-            "create a new empty PDF from stratch"
+            "Create a new empty PDF from stratch."
         )
         .def_static("open", open_pdf,
             R"~~~(
-            Open an existing file at `filename_or_stream`.
+            Open an existing file at *filename_or_stream*.
 
-            If `filename_or_stream` is path-like, the file will be opened. The
-            file should not be modified by another process while it is open in
-            pikepdf.
+            If *filename_or_stream* is path-like, the file will be opened for reading.
+            The file should not be modified by another process while it is open in
+            pikepdf. The file will not be altered when opened in this way. Any changes
+            to the file must be persisted by using ``.save()``.
 
-            If `filename_or_stream` has `.read()` and `.seek()` methods, the file
+            If *filename_or_stream* has ``.read()`` and ``.seek()`` methods, the file
             will be accessed as a readable binary stream. pikepdf will read the
             entire stream into a private buffer.
 
+            ``.open()`` may be used in a ``with``-block, ``.close()`` will be called when
+            the block exists.
+
+            Examples:
+
+                >>> with Pdf.open("test.pdf") as pdf:
+                        ...
+
+                >>> pdf = Pdf.open("test.pdf", password="rosebud")
+
             Args:
                 filename_or_stream (os.PathLike): Filename of PDF to open
                 password (str or bytes): User or owner password to open an
@@ -278,7 +499,8 @@ PYBIND11_MODULE(_qpdf, m) {
                 ignore_xref_streams (bool): If True, ignore cross-reference
                     streams. See qpdf documentation.
                 suppress_warnings (bool): If True (default), warnings are not
-                    printed to stderr. Use `get_warnings()` to retrieve warnings.
+                    printed to stderr. Use :meth:`pikepdf.Pdf.get_warnings()` to
+                    retrieve warnings.
                 attempt_recovery (bool): If True (default), attempt to recover
                     from PDF parsing errors.
                 inherit_page_attributes (bool): If True (default), push attributes
@@ -289,7 +511,7 @@ PYBIND11_MODULE(_qpdf, m) {
                     file.
                 pikepdf.PdfError: If for other reasons we could not open
                     the file.
-                TypeError: If the type of `filename_or_stream` is not
+                TypeError: If the type of ``filename_or_stream`` is not
                     usable.
                 FileNotFoundError: If the file was not found.
             )~~~",
@@ -307,15 +529,15 @@ PYBIND11_MODULE(_qpdf, m) {
             }
         )
         .def_property_readonly("filename", &QPDF::getFilename,
-            "the source filename of an existing PDF, when available")
+            "The source filename of an existing PDF, when available.")
         .def_property_readonly("pdf_version", &QPDF::getPDFVersion,
-            "the PDF standard version, such as '1.7'")
+            "The version of the PDF specification used for this file, such as '1.7'.")
         .def_property_readonly("extension_level", &QPDF::getExtensionLevel)
         .def_property_readonly("Root", &QPDF::getRoot,
-            "the /Root object of the PDF"
+            "The /Root object of the PDF."
         )
         .def_property_readonly("root", &QPDF::getRoot,
-            "alias for .Root, the /Root object of the PDF"
+            "Alias for .Root, the /Root object of the PDF."
         )
         .def_property("docinfo",
             [](QPDF& q) {
@@ -330,7 +552,16 @@ PYBIND11_MODULE(_qpdf, m) {
                     throw py::value_error("docinfo must be an indirect object - use Pdf.make_indirect");
                 q.getTrailer().replaceKey("/Info", replace);
             },
-            "access the document information dictionary"
+            R"~~~(
+            Access the (deprecated) document information dictionary.
+
+            The document information dictionary is a brief metadata record
+            that can store some information about the origin of a PDF. It is
+            deprecated and removed in the PDF 2.0 specification. Use the
+            ``.open_metadata()`` API instead, which will edit the modern (and
+            unfortunately, more complicated) XMP metadata object and synchronize
+            changes to the document information dictionary.
+            )~~~"
         )
         .def_property_readonly("trailer", &QPDF::getTrailer,
             R"~~~(
@@ -346,6 +577,12 @@ PYBIND11_MODULE(_qpdf, m) {
             [](std::shared_ptr<QPDF> q) {
                 return PageList(q);
             },
+            R"~~~(
+            Returns the list of pages.
+
+            Return type:
+                pikepdf._qpdf.PageList
+            )~~~",
             py::keep_alive<0, 1>()
         )
         .def_property_readonly("_pages", &QPDF::getAllPages)
@@ -377,7 +614,15 @@ PYBIND11_MODULE(_qpdf, m) {
             )~~~",
             py::arg_v("stream", py::module::import("sys").attr("stderr"), "sys.stderr")
         )
-        .def("get_warnings", &QPDF::getWarnings)  // this is a def because it modifies state by clearing warnings
+        .def("get_warnings", // this is a def because it modifies state by clearing warnings
+            [](QPDF& q) {
+                py::list warnings;
+                for (auto w: q.getWarnings()) {
+                    warnings.append(w.what());
+                }
+                return warnings;
+            }
+        )
         .def("show_xref_table", &QPDF::showXRefTable,
             R"~~~(
             Pretty-print the Pdf's xref (cross-reference table)
@@ -394,8 +639,9 @@ PYBIND11_MODULE(_qpdf, m) {
             The page can be either be a newly constructed PDF object or it can
             be obtained from another PDF.
 
-            :param pikepdf.Object page: The page object to attach
-            :param bool first: If True, prepend this before the first page; if False append after last page
+            Args:
+                page (pikepdf.Object): The page object to attach
+                first (bool): If True, prepend this before the first page; if False append after last page
             )~~~",
             py::arg("page"),
             py::arg("first")=false,
@@ -423,11 +669,14 @@ PYBIND11_MODULE(_qpdf, m) {
         .def("save",
             save_pdf,
             R"~~~(
-            Save all modifications to this :class:`pikepdf.Pdf`
+            Save all modifications to this :class:`pikepdf.Pdf`.
 
             Args:
                 filename (str or stream): Where to write the output. If a file
-                    exists in this location it will be overwritten.
+                    exists in this location it will be overwritten. The file
+                    should not be the same as the input file, because data from
+                    the input file may be lazily loaded; as such overwriting
+                    in place will null-out objects.
 
                 static_id (bool): Indicates that the ``/ID`` metadata, normally
                     calculated as a hash of certain PDF contents and metadata
@@ -437,12 +686,21 @@ PYBIND11_MODULE(_qpdf, m) {
                     manner compliant with PDF/A and other stricter variants.
                     This should be True, the default, in most cases.
 
-                min_version (str): Sets the minimum version of PDF
+                min_version (str or tuple): Sets the minimum version of PDF
                     specification that should be required. If left alone QPDF
-                    will decide.
-                force_version (str): Override the version recommend by QPDF,
+                    will decide. If a tuple, the second element is an integer, the
+                    extension level. If the version number is not a valid format,
+                    QPDF will decide what to do.
+                force_version (str or tuple): Override the version recommend by QPDF,
                     potentially creating an invalid file that does not display
-                    in old versions. See QPDF manual for details.
+                    in old versions. See QPDF manual for details. If a tuple, the
+                    second element is an integer, the extension level.
+                fix_metadata_version (bool): If True (default) and the XMP metadata
+                    contains the optional PDF version field, ensure the version in
+                    metadata is correct. If the XMP metadata does not contain a PDF
+                    version field, none will be added. To ensure that the field is
+                    added, edit the metadata and insert a placeholder value in
+                    ``pdf:PDFVersion``.
 
                 object_stream_mode (pikepdf.ObjectStreamMode):
                     ``disable`` prevents the use of object streams.
@@ -472,10 +730,24 @@ PYBIND11_MODULE(_qpdf, m) {
                     the program ``fix-qdf`` to fix convert back to a standard
                     PDF.
 
+                progress (callable): Specify a callback function that is called
+                    as the PDF is written. The function will be called with an
+                    integer between 0-100 as the sole parameter, the progress
+                    percentage. This function may not access or modify the PDF
+                    while it is being written, or data corruption will almost
+                    certainly occur.
+
+                encryption (pikepdf.models.Encryption or bool): If ``False``
+                    or omitted, existing encryption will be removed. If ``True``
+                    encryption settings are copied from the originating PDF.
+                    Alternately, an ``Encryption`` object may be provided that
+                    sets the parameters for new encryption.
+
             You may call ``.save()`` multiple times with different parameters
             to generate different versions of a file, and you *may* continue
             to modify the file after saving it. ``.save()`` does not modify
-            the ``Pdf`` object in memory.
+            the ``Pdf`` object in memory, except possibly by updating the XMP
+            metadata version with ``fix_metadata_version``.
 
             .. note::
 
@@ -485,19 +757,26 @@ PYBIND11_MODULE(_qpdf, m) {
                 done automatically because ``.save()`` is intended to be
                 idempotent.
 
+            .. note::
+
+                pikepdf can read PDFs will incremental updates, but always
+                any coalesces incremental updates into a single non-incremental
+                PDF file when saving.
             )~~~",
             py::arg("filename"),
             py::arg("static_id")=false,
             py::arg("preserve_pdfa")=true,
             py::arg("min_version")="",
             py::arg("force_version")="",
+            py::arg("fix_metadata_version")=true,
             py::arg("compress_streams")=true,
-            py::arg("stream_decode_level")=qpdf_stream_decode_level_e::qpdf_dl_generalized,
+            py::arg("stream_decode_level")=py::none(),
             py::arg("object_stream_mode")=qpdf_object_stream_e::qpdf_o_preserve,
             py::arg("normalize_content")=false,
             py::arg("linearize")=false,
             py::arg("qdf")=false,
-            py::arg("progress")=py::none()
+            py::arg("progress")=py::none(),
+            py::arg("encryption")=py::none()
         )
         .def("_get_object_id", &QPDF::getObjectByID)
         .def("get_object",
@@ -507,10 +786,11 @@ PYBIND11_MODULE(_qpdf, m) {
             R"~~~(
             Look up an object by ID and generation number
 
-            Returns:
+            Return type:
                 pikepdf.Object
             )~~~",
-            py::return_value_policy::reference_internal
+            py::return_value_policy::reference_internal,
+            py::arg("objgen")
         )
         .def("get_object",
             [](QPDF &q, int objid, int gen) {
@@ -519,10 +799,12 @@ PYBIND11_MODULE(_qpdf, m) {
             R"~~~(
             Look up an object by ID and generation number
 
-            Returns:
+            Return type:
                 pikepdf.Object
             )~~~",
-            py::return_value_policy::reference_internal
+            py::return_value_policy::reference_internal,
+            py::arg("objid"),
+            py::arg("gen")
         )
         .def("make_indirect", &QPDF::makeIndirectObject,
             R"~~~(
@@ -542,9 +824,10 @@ PYBIND11_MODULE(_qpdf, m) {
             See Also:
                 :meth:`pikepdf.Object.is_indirect`
 
-            Returns:
+            Return type:
                 pikepdf.Object
-            )~~~"
+            )~~~",
+            py::arg("h")
         )
         .def("make_indirect",
             [](QPDF &q, py::object obj) -> QPDFObjectHandle {
@@ -553,9 +836,10 @@ PYBIND11_MODULE(_qpdf, m) {
             R"~~~(
             Encode a Python object and attach to this Pdf as an indirect object
 
-            Returns:
+            Return type:
                 pikepdf.Object
-            )~~~"
+            )~~~",
+            py::arg("obj")
         )
         .def("copy_foreign",
             [](QPDF &q, QPDFObjectHandle &h) -> QPDFObjectHandle {
@@ -563,20 +847,106 @@ PYBIND11_MODULE(_qpdf, m) {
             },
             "Copy object from foreign PDF to this one.",
             py::return_value_policy::reference_internal,
-            py::keep_alive<1, 2>()
+            py::keep_alive<1, 2>(),
+            py::arg("h")
         )
         .def("_replace_object",
             [](QPDF &q, int objid, int gen, QPDFObjectHandle &h) {
                 q.replaceObject(objid, gen, h);
             }
         )
+        .def("_swap_objects",
+            [](QPDF &q, std::pair<int, int> objgen1, std::pair<int, int> objgen2) {
+                QPDFObjGen o1(objgen1.first, objgen1.second);
+                QPDFObjGen o2(objgen2.first, objgen2.second);
+                q.swapObjects(o1, o2);
+            }
+        )
+        .def("_process",
+            [](QPDF &q, std::string description, py::bytes data) {
+                std::string s = data;
+                q.processMemoryFile(
+                    description.c_str(),
+                    s.data(),
+                    s.size()
+                );
+            },
+            R"~~~(
+            Process a new in-memory PDF, replacing the existing PDF
+
+            Used to implement Pdf.close().
+            )~~~"
+        )
+        .def_property_readonly("_allow_accessibility",
+            [](QPDF &q) {
+                return q.allowAccessibility();
+            }
+        )
+        .def_property_readonly("_allow_extract",
+            [](QPDF &q) {
+                return q.allowExtractAll();
+            }
+        )
+        .def_property_readonly("_allow_print_lowres",
+            [](QPDF &q) {
+                return q.allowPrintLowRes();
+            }
+        )
+        .def_property_readonly("_allow_print_highres",
+            [](QPDF &q) {
+                return q.allowPrintHighRes();
+            }
+        )
+        .def_property_readonly("_allow_modify_assembly",
+            [](QPDF &q) {
+                return q.allowModifyAssembly();
+            }
+        )
+        .def_property_readonly("_allow_modify_form",
+            [](QPDF &q) {
+                return q.allowModifyForm();
+            }
+        )
+        .def_property_readonly("_allow_modify_annotation",
+            [](QPDF &q) {
+                return q.allowModifyAnnotation();
+            }
+        )
+        .def_property_readonly("_allow_modify_other",
+            [](QPDF &q) {
+                return q.allowModifyOther();
+            }
+        )
+        .def_property_readonly("_allow_modify_all",
+            [](QPDF &q) {
+                return q.allowModifyAll();
+            }
+        )
+        .def_property_readonly("_encryption_data",
+            [](QPDF &q) {
+                int R = 0;
+                int P = 0;
+                int V = 0;
+                QPDF::encryption_method_e stream_method = QPDF::e_unknown;
+                QPDF::encryption_method_e string_method = QPDF::e_unknown;
+                QPDF::encryption_method_e file_method = QPDF::e_unknown;
+                if (!q.isEncrypted(R, P, V, stream_method, string_method, file_method))
+                    return py::dict();
+
+                auto user_passwd = q.getTrimmedUserPassword();
+                auto encryption_key = q.getEncryptionKey();
+
+                return py::dict(
+                    py::arg("R") = R,
+                    py::arg("P") = P,
+                    py::arg("V") = V,
+                    py::arg("stream") = stream_method,
+                    py::arg("string") = string_method,
+                    py::arg("file") = file_method,
+                    py::arg("user_passwd") = py::bytes(user_passwd),
+                    py::arg("encryption_key") = py::bytes(encryption_key)
+                );
+            }
+        )
         ; // class Pdf
-
-    init_object(m);
-
-#ifdef VERSION_INFO
-    m.attr("__version__") = VERSION_INFO;
-#else
-    m.attr("__version__") = "dev";
-#endif
 }
