@@ -121,18 +121,6 @@ void PageList::set_pages_from_iterable(py::slice slice, py::iterable other)
 void PageList::delete_page(size_t index)
 {
     auto page = this->get_page(index);
-    /*
-    // Need a dec_ref to match the inc_ref in insert_page, but it's unclear
-    // how to do that. The item will be set the current QPDF always.
-    // Accessing data from another PDF seems to involve some pipeline
-    // magic in QPDF around libqpdf/QPDFWriter.cc:1614
-    if (original page owner != &this->getQPDF()) {
-        // If we are removing a page not originally owned by our QPDF,
-        // remove the reference count we put it in insert_page()
-        py::object pyqpdf = py::cast(page_owner);
-        pyqpdf.dec_ref();
-    }
-    */
     this->qpdf->removePage(page);
 }
 
@@ -175,24 +163,6 @@ void PageList::insert_page(size_t index, QPDFObjectHandle page)
         // qpdf does not accept duplicating pages within the same file,
         // so manually create a copy
         page = this->qpdf->makeIndirectObject(page);
-    } else {
-        // libqpdf does not transfer a page's contents to the new QPDF.
-        // Instead WHEN ASKED TO WRITE it will go back and get the data
-        // from objecthandle->getOwningQPDF(). Therefore we must ensure
-        // our previous owner is kept alive.
-#if 1
-        auto tinfo = py::detail::get_type_info(typeid(QPDF));
-        py::handle pyqpdf = py::detail::get_object_handle(page_owner, tinfo);
-        py::handle pypage = py::cast(page);
-        py::detail::keep_alive_impl(pypage, pyqpdf);
-#else
-        // MSVC++ complains about the symbol
-        // QPDF::Members::~Members() not being exported when this version
-        // is used, but it works for GCC and Clang.
-        py::handle pyqpdf = py::cast(page_owner);
-        py::handle pypage = py::cast(page);
-        py::detail::keep_alive_impl(pypage, pyqpdf);
-#endif
     }
     if (index != this->count()) {
         QPDFObjectHandle refpage = this->get_page(index);
@@ -202,7 +172,6 @@ void PageList::insert_page(size_t index, QPDFObjectHandle page)
     }
 }
 
-
 void init_pagelist(py::module &m)
 {
     py::class_<PageList>(m, "PageList")
@@ -229,12 +198,13 @@ void init_pagelist(py::module &m)
         .def("__delitem__", &PageList::delete_pages_from_iterable)
         .def("__len__", &PageList::count)
         .def("p",
-            [](PageList &pl, size_t index) {
-                if (index == 0) // Indexing past end is checked in .get_page
+            [](PageList &pl, size_t pnum) {
+                if (pnum == 0) // Indexing past end is checked in .get_page
                     throw py::index_error("page access out of range in 1-based indexing");
-                return pl.get_page(index - 1);
+                return pl.get_page(pnum - 1);
             },
-            "convenience - look up page number in ordinal numbering, .p(1) is first page"
+            "Convenience - look up page number in ordinal numbering, ``.p(1)`` is first page",
+            py::arg("pnum")
         )
         .def("__iter__",
             [](PageList &pl) {
@@ -252,7 +222,16 @@ void init_pagelist(py::module &m)
             [](PageList &pl, ssize_t index, py::object obj) {
                 size_t uindex = uindex_from_index(pl, index);
                 pl.insert_page(uindex, obj);
-            }, py::keep_alive<1, 3>()
+            }, py::keep_alive<1, 3>(),
+            R"~~~(
+            Insert a page at the specified location.
+
+            Args:
+                index (int): location at which to insert page, 0-based indexing
+                obj (pikepdf.Object): page object to insert
+            )~~~",
+            py::arg("index"),
+            py::arg("obj")
         )
         .def("reverse",
             [](PageList &pl) {
@@ -262,13 +241,16 @@ void init_pagelist(py::module &m)
                     PySlice_New(Py_None, Py_None, step.ptr()));
                 py::list reversed_pages = pl.get_pages(reversed);
                 pl.set_pages_from_iterable(ordinary_indices, reversed_pages);
-            }
+            },
+            "Reverse the order of pages."
         )
         .def("append",
             [](PageList &pl, py::object page) {
                 pl.insert_page(pl.count(), page);
             },
-            py::keep_alive<1, 2>()
+            py::keep_alive<1, 2>(),
+            "Add another page to the end.",
+            py::arg("page")
         )
         .def("extend",
             [](PageList &pl, PageList &other) {
@@ -279,7 +261,9 @@ void init_pagelist(py::module &m)
                     pl.insert_page(pl.count(), other.get_page(i));
                 }
             },
-            py::keep_alive<1, 2>()
+            py::keep_alive<1, 2>(),
+            "Extend the ``Pdf`` by adding pages from another ``Pdf.pages``.",
+            py::arg("other")
         )
         .def("extend",
             [](PageList &pl, py::iterable iterable) {
@@ -290,6 +274,29 @@ void init_pagelist(py::module &m)
                     ++it;
                 }
             },
-            py::keep_alive<1, 2>()
+            py::keep_alive<1, 2>(),
+            "Extend the ``Pdf`` by adding pages from an iterable of pages.",
+            py::arg("iterable")
+        )
+        .def("remove",
+            [](PageList &pl, py::kwargs kwargs) {
+                auto pnum = kwargs["p"].cast<size_t>();
+                if (pnum == 0) // Indexing past end is checked in .get_page
+                    throw py::index_error("page access out of range in 1-based indexing");
+                pl.delete_page(pnum - 1);
+            },
+            R"~~~(
+            Remove a page (using 1-based numbering)
+
+            Args:
+                p (int): 1-based page number
+            )~~~"
+        )
+        .def("__repr__",
+            [](PageList &pl) {
+                return std::string("<pikepdf._qpdf.PageList len=")
+                    + std::to_string(pl.count())
+                    + std::string(">");
+            }
         );
 }
