@@ -1,12 +1,16 @@
-import pytest
-from pikepdf import Pdf, Stream, PdfMatrix
-
+import gc
 from contextlib import suppress
 from shutil import copy
-import gc
-
 from sys import getrefcount as refcount
 
+import pytest
+
+from pikepdf import Pdf, PdfMatrix, Stream
+
+
+# pylint: disable=redefined-outer-name,pointless-statement
+
+
 @pytest.fixture
 def graph(resources):
     return Pdf.open(resources / 'graph.pdf')
@@ -47,12 +51,17 @@ def test_delete_last_page(graph, outdir):
 def test_replace_page(graph, fourpages):
     q = fourpages
     q2 = graph
+    q2.pages[0].CropBox = [0, 0, 500, 500]
+
+    # Ensure the page keys are different, not subsets
+    assert q.pages[1].keys() - q2.pages[0].keys()
+    assert q2.pages[0].keys() - q.pages[1].keys()
 
     assert len(q.pages) == 4
     q.pages[1] = q2.pages[0]
     assert len(q.pages) == 4
-    assert q.pages[1].Resources.XObject.keys() == \
-        q2.pages[0].Resources.XObject.keys()
+    assert q.pages[1].keys() == q2.pages[0].keys()
+    assert q.pages[1].Resources.XObject.keys() == q2.pages[0].Resources.XObject.keys()
 
 
 def test_hard_replace_page(fourpages, graph, sandwich, outdir):
@@ -98,11 +107,11 @@ def test_evil_page_deletion(resources, outdir):
 
     assert refcount(src) == 2
     pdf.pages.append(src.pages[0])
-    assert refcount(src) == 3
+    assert refcount(src) == 2
 
     del src.pages[0]
     gc.collect()
-    assert refcount(src) == 3
+    assert refcount(src) == 2
 
     with suppress(PermissionError):  # Fails on Windows
         (outdir / 'sandwich.pdf').unlink()
@@ -115,9 +124,6 @@ def test_evil_page_deletion(resources, outdir):
     pdf.save(outdir / 'out_nopages.pdf')
     del pdf
     gc.collect()
-    # Ideally we'd see the check_refcount(src, 2) at this point, but we don't
-    # have a way to find out when a PDF can be closed if a page was copied out
-    # of it to another PDF
 
 
 def test_append_all(sandwich, fourpages, outdir):
@@ -154,10 +160,12 @@ def test_slice_unequal_replacement(fourpages, sandwich, outdir):
 
     assert len(pdf.pages) == 2, "number of pages must be changed"
     pdf.save(outdir / 'out.pdf')
-    assert pdf.pages[0].Contents.Length == page0_content_len, \
-        "page 0 should be unchanged"
-    assert pdf.pages[1].Contents.Length != page1_content_len, \
-        "page 1's contents should have changed"
+    assert (
+        pdf.pages[0].Contents.Length == page0_content_len
+    ), "page 0 should be unchanged"
+    assert (
+        pdf.pages[1].Contents.Length != page1_content_len
+    ), "page 1's contents should have changed"
 
 
 def test_slice_with_step(fourpages, sandwich, outdir):
@@ -171,24 +179,21 @@ def test_slice_with_step(fourpages, sandwich, outdir):
     pdf.pages[0::2] = pdf2.pages
     pdf.save(outdir / 'out.pdf')
 
-    assert all(page.Contents.Length == pdf2_content_len
-               for page in pdf.pages[0::2])
+    assert all(page.Contents.Length == pdf2_content_len for page in pdf.pages[0::2])
 
 
 def test_slice_differing_lengths(fourpages, sandwich):
     pdf = fourpages
     pdf2 = sandwich
 
-    with pytest.raises(ValueError,
-            message="attempt to assign"):
+    with pytest.raises(ValueError, match="attempt to assign"):
         pdf.pages[0::2] = pdf2.pages[0:1]
 
 
 @pytest.mark.timeout(1)
 def test_self_extend(fourpages):
     pdf = fourpages
-    with pytest.raises(ValueError,
-            message="source page list modified during iteration"):
+    with pytest.raises(ValueError, match="source page list modified during iteration"):
         pdf.pages.extend(pdf.pages)
 
 
@@ -240,3 +245,26 @@ def test_negative_indexing(fourpages, graph):
         fourpages.pages[-42] = graph.pages[0]
     with pytest.raises(IndexError):
         del fourpages.pages[-42]
+
+
+def test_concatenate(resources, outdir):
+    # Issue #22
+    def concatenate(n):
+        print('concatenating same page', n, 'times')
+        output_pdf = Pdf.new()
+        for i in range(n):
+            print(i)
+            pdf_page = Pdf.open(resources / 'pal.pdf')
+            output_pdf.pages.extend(pdf_page.pages)
+        output_pdf.save(outdir / '{}.pdf'.format(n))
+
+    concatenate(5)
+
+
+def test_emplace(fourpages):
+    p0_objgen = fourpages.pages[0].objgen
+    fourpages.pages[0].emplace(fourpages.pages[1])
+    assert p0_objgen == fourpages.pages[0].objgen
+    assert fourpages.pages[0].keys() == fourpages.pages[1].keys()
+    for k in fourpages.pages[0].keys():
+        assert fourpages.pages[0][k] == fourpages.pages[1][k]
